"""Merge a base class with one or more extension classes."""

from typing import Any, Callable, Dict, List, Tuple, Type
from collections import deque
import inspect

from dyndesign.dynloader import importclass

__all__ = ["mergeclasses"]


def __adapt_arguments(func: Callable, *args, **kwargs) -> Tuple[List, Dict]:
    """Filter `args` and `kwargs` based and the arguments accepted by an input function.

    :param func: input function.
    :param args: input arguments.
    :param kwargs: input keyword arguments.
    :return: filtered arguments and keyword arguments.
    """
    init_specs = inspect.getfullargspec(func)
    if init_specs.varargs:
        res_args = list(args)
    else:
        func_args = init_specs.args[1:]
        arg_deque = deque(args)
        res_args = []
        for func_arg in func_args:
            if func_arg in kwargs:
                res_args.append(kwargs[func_arg])
                del kwargs[func_arg]
            elif arg_deque:
                res_args.append(arg_deque.popleft())

    func_kwargs = init_specs.kwonlydefaults or {}
    if init_specs.varkw:
        res_kwargs = kwargs
    else:
        res_kwargs = {key: value for key, value in kwargs.items() if key in func_kwargs}

    return res_args, res_kwargs


def __merge_class_inits(classes: List[Type]) -> Callable:
    """Build a merged constructor by calling the constructors of the merged classes.

    :param classes: merged classes.
    :return: merged constructor.
    """
    class_constructors = []
    for cur_class in classes:
        if '__init__' in dir(cur_class):
            class_constructors.append(cur_class.__init__)

    def init_all_classes(obj, *args, **kwargs):
        for class_constructor in class_constructors:
            if not inspect.ismethoddescriptor(class_constructor):
                filtered_args, filtered_kwargs = __adapt_arguments(class_constructor, *args, **kwargs)
                class_constructor(obj, *filtered_args, **filtered_kwargs)

    return init_all_classes


def __preprocess_classes(all_classes: Any) -> List[Type]:
    """Dynamically import classes if passed as strings."""
    result_classes = []
    for class_id in all_classes:
        if type(class_id) == str:
            cl = importclass(class_id)
        else:
            cl = class_id
        result_classes.append(cl)
    return result_classes


def mergeclasses(base_class: Any, *extension_classes: Any) -> Type:
    """Merge (i.e., extend) a base class with one or more extension classes. If more than one adapter classes are
    provided, then the classes are extended in sequence (from the first one to the last).

    :param base_class: base class.
    :param extension_classes: extension classes.
    :return: merged class.
    """
    all_classes = __preprocess_classes((base_class,) + extension_classes)
    init_all_classes = __merge_class_inits(all_classes)
    return type(
        all_classes[0].__name__,
        tuple(all_classes[::-1]),
        {"__init__": init_all_classes}
    )



"""Import dynamically a class."""

from typing import Type

__all__ = ["importclass"]


def importclass(
    module_name: str,
    class_name: str = None  # type: ignore
) -> Type:
    """Dynamically import a class from a python module.

    :param module_name: name of the module to import.
    :param class_name: name of the class in the module to import, if different from the module's name.
    :return: dynamically imported class.
    """
    if not class_name:
        module_name, class_name = module_name.rsplit('.', 1)
    loaded_module = __import__(module_name, fromlist=class_name)
    return getattr(loaded_module, class_name)



"""Invoke methods that have been dynamically added to a class."""

from contextlib import AbstractContextManager
from functools import wraps
from operator import attrgetter
import re
from typing import Any, Callable, List

__all__ = ["decoratewith", "safeinvoke", "safezone"]


def __is_sub_object(method_name: str):
    return '.' in method_name


def __try_invoke_method(
    method_name: str,
    instance: object,
    *args,
    **kwargs
) -> Any:
    """Attempt to invoke a method of a class instance.

    :param method_name: name of a method of the class instance or path (in dot notation) to a method of a
                        sub-instance.
    :param instance: class instance that may optionally include the method referenced to with `method_name`.
    :return: value returned by the method, if such a method exists.
    """
    try:
        if __is_sub_object(method_name):
            method = attrgetter(method_name)(instance)
        else:
            method = getattr(instance, method_name)
    except AttributeError:
        raise ErrorMethodNotFound
    return method.__call__(*args, **kwargs)


def decoratewith(
    *method_name_args: str,
    method_sub_instance: str = None,  # type: ignore
    fallback: Callable = None  # type: ignore
) -> Callable:
    """Meta decorator to decorate a function with one or more decorator methods. Decorator methods can be, for
    example, class methods, methods dynamically added and/or methods of any sub-instance of a class instance.

    :param method_name_args: method name(s) of the one or more decorator methods.
    :param method_sub_instance: name of the sub-instance to be prepended to each method name.
    :return: method decorator.
    """
    method_names=list(method_name_args)
    if method_sub_instance:
        method_names = [f"{method_sub_instance}.{m}" for m in method_names]

    def dynamic_decorator_wrapper(
        func: Callable,
        method_names: List[str] = method_names
    ) -> Callable:
        """Decorator wrapper using one or more decorator methods in pipeline to decorate a function `func`.

        :param func: function to decorate.
        :param method_names: method name(s) of the one or more decorator methods.
        :return: decorated method.
        """
        method_name = method_names.pop()

        @wraps(func)
        def dynamic_decorator_func(instance, *args, **kwargs) -> Any:
            decorator_args = (func, ) + args
            if __is_sub_object(method_name):
                kwargs["decorated_self"] = instance
            try:
                return __try_invoke_method(method_name, instance, *decorator_args, **kwargs)
            except ErrorMethodNotFound:
                kwargs.pop("decorated_self", None)
                if fallback:
                    fallback(instance, *args, **kwargs)
                return func(instance, *args, **kwargs)

        if method_names:
            return dynamic_decorator_wrapper(dynamic_decorator_func, method_names=method_names)
        else:
            return dynamic_decorator_func

    return dynamic_decorator_wrapper


def safeinvoke(
    method_name: str,
    instance: object,
    fallback: Callable = None,  # type: ignore
    *args,
    **kwargs
) -> Any:
    """Attempt to invoke a method of a class instance. If the method is not found, code execution proceeds
    normally.

    :param method_name: name of a method of the class instance or path (in dot notation) to a method of a
                        sub-instance.
    :param instance: class instance that may optionally include the method referenced to with `method_name`.
    :return: value returned by the method, if such a method exists.
    """
    try:
        return __try_invoke_method(method_name, instance, *args, **kwargs)
    except ErrorMethodNotFound:
        if fallback:
            fallback(*args, **kwargs)
        return None


class safezone(AbstractContextManager):
    """Context manager to suppress `AttributeError` and `NameError` exceptions raised when specific methods are not
    available. After suppressing the exception, the execution proceeds with the next statement following the `with`
    statement.
    """

    def __init__(self,
        *method_names: str,
        fallback: Callable = None,  # type: ignore
    ):
        self.__method_names = method_names
        self.__fallback = fallback


    def __enter__(self):
        pass


    def __is_protected_name(self, excinst) -> bool:
        if not self.__method_names:
            return True
        try:
            method_name = excinst.name
        except AttributeError:
            method_name = re.findall(r"'([^']+)'", excinst.args[0])[-1]
        return method_name in self.__method_names


    def __exit__(self, exctype, excinst, exctb) -> bool:
        expected_exception = AttributeError if (
            'tb_frame' in dir(exctb) and
            'self' in exctb.tb_frame.f_locals
        ) else NameError
        result = (
            exctype is not None and
            issubclass(exctype, expected_exception) and
            self.__is_protected_name(excinst)
        )
        if result and self.__fallback:
            self.__fallback()
        return result


class ErrorMethodNotFound(Exception):
    """Raised when a dynamic method cannot be found."""



"""Create and destroy Singleton classes."""

from typing import Any

__all__ = ["SingletonMeta"]


class SingletonMeta(type):
    """Meta class to instantiate Singleton classes."""

    _instances: Any = {}

    def __new__(cls, name, bases, dct) -> type:
        """Add the class method `destroy` to the Singleton class."""
        def __destroy_singleton(_):
            cls.destroy(name)
        instance = super().__new__(cls, name, bases, dct)
        setattr(instance, 'destroy_singleton', __destroy_singleton)
        return instance


    def __call__(cls, *args, **kwargs) -> type:
        """Return the Singleton class instance, if any instance is found, otherwise create and return a new Singleton
        class instance.

        :return: Singleton class instance.
        """
        if cls not in cls._instances:
            instance = super(SingletonMeta, cls).__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


    @classmethod
    def destroy(cls, *class_names: str):
        """Delete all the instances of the Singleton classes whose names are passed as arguments. If no class name is
        passed, the instances of all the Singleton classes are deleted.

        :param class_names: names of the Singleton class instances to destroy.
        """
        if class_names:
            cls._instances = dict(filter(lambda k: k[0].__name__ not in class_names, cls._instances.items()))
        else:
            cls._instances = {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dyndesign']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dyndesign',
    'version': '0.9.2',
    'description': 'Toolset for Dynamic Design Patterns in Python.',
    'long_description': 'DynDesign\n=========\n\n|Build Status| |PyPi Version Status| |Python Version Status| |License|\n\nA set of tools for Dynamic Design Patterns in Python.\n\n\nInstall\n-------\n\nDyndesign is on the Python Package Index (PyPI):\n\n::\n\n    pip install dyndesign\n\n\nOverview\n--------\n\nMerge two or more classes:\n\n.. code:: python\n\n    from dyndesign import mergeclasses\n\n    MergedClass = mergeclasses(Base, Ext1, Ext2, ...)\n\nDecorate a method with one or more instance methods loaded at runtime:\n\n.. code:: python\n\n    from dyndesign import decoratewith\n\n    @decoratewith("decorator_1", "component.decorator_2", ...)\n    def decorated_method(self, ...):\n        ...\n\nSafely invoke functions or methods from a ``safezone`` context manager:\n\n.. code:: python\n\n    from dyndesign import safezone\n\n    with safezone():\n        ...\n        function_possibly_non_existent()\n        ...\n\nCreate and destroy Singleton classes:\n\n.. code:: python\n\n    from dyndesign import SingletonMeta\n\n    class Singleton(metaclass=SingletonMeta):\n        ...\n\n    singleton_instance = Singleton(...)\n    same_singleton_instance = Singleton()\n    Singleton().destroy_singleton()\n    new_singleton_instance = Singleton(...)\n\nImport classes dynamically using the path:\n\n.. code:: python\n\n    from dyndesign import importclass\n\n    ImportedClass = importclass("directory.module.class_name")\n\n\nMerging Classes\n---------------\n\nDyndesign provides API ``mergeclasses`` to merge two or more classes as if they\nwere dictionaries, so that the merged class has the attributes and methods of\nthe base class and of the extension classes. If two or more classes have the\nsame attributes/methods, the attributes/methods of the rightmost classes (in the\norder in which they are passed to ``mergeclasses``) overwrite the leftmost,\nsimilarly to what happens when merging dictionaries.\n\n.. code:: python\n\n    from dyndesign import mergeclasses\n\n    class Base:\n        def __init__(self, init_value):\n            self.param = init_value\n\n        def m1(self):\n            print(f"Method `m1` of class `Base`, and {self.param=}")\n\n        def m2(self):\n            print(f"Method `m2` of class `Base`")\n\n    class Ext:\n        def m1(self):\n            print(f"Method `m1` of class `Ext`, and {self.param=}")\n\n    MergedClass = mergeclasses(Base, Ext)\n    merged_instance = MergedClass("INITIAL VALUE")\n    merged_instance.m1()\n    merged_instance.m2()\n\n    # Method `m1` of class `Ext`, and self.param=\'INITIAL VALUE\'\n    # Method `m2` of class `Base`\n\n\nWhen a merged class is instantiated with arguments, the constructor of each\nmerging class takes just the arguments it needs (i.e., the arguments in its\nsignature):\n\n.. code:: python\n\n    from dyndesign import mergeclasses\n\n    class A:\n        def __init__(self):\n            print("No argument passed to class `A`")\n\n    class B:\n        def __init__(self, a):\n            print(f"Argument {a=} passed to class `B`")\n\n    class C:\n        def __init__(self, a, b, kw1=None):\n            print(f"Argument {a=}, {b=} and {kw1=} passed to class `C`")\n\n    class D:\n        def __init__(self, kw2=None):\n            print(f"Argument {kw2=} passed to class `D`")\n\n    MergedClass = mergeclasses(A, B, C, D)\n    MergedClass("Alpha", "Beta", kw1="kwarg #1", kw2="kwarg #2")\n\n    # No argument passed to class `A`\n    # Argument a=\'Alpha\' passed to class `B`\n    # Argument a=\'Alpha\', b=\'Beta\' and kw1=\'kwarg #1\' passed to class `C`\n    # Argument kw2=\'kwarg #2\' passed to class `D`\n\n\nDynamic Decorators\n------------------\n\nMeta decorator ``decoratewith`` decorates a class method with one or more\npipelined instance decorators (regardless whether they statically exist or not).\nThe syntax of the dynamic decorators aims to get rid of the boilerplate for\nwrapping and returning the decorator code, leaving just the wrapper\'s code. For\nexample, dynamic decorators can be used to decorate a method from a base\nclass with a method from an extension class:\n\n.. code:: python\n\n    from dyndesign import decoratewith\n\n    class Base:\n        @decoratewith("decorator")\n        def m(self):\n            print(f"Method `m` of class `Base`")\n\n    class Ext:\n        def decorator(self, func):\n            print("Beginning of method decoration.")\n            func(self)\n            print("End of method decoration.")\n\n    merged = mergeclasses(Base, Ext)()\n    merged.m()\n\n    # Beginning of method decoration.\n    # Method `m` of class `Base`\n    # End of method decoration.\n\n\nArguments of ``decoratewith`` are loaded at runtime as properties of the\nvariable \'self\': a dynamic decorator can be, for example, a method of a\ncomponent class. In case of dynamic decoration from a sub-instance of \'self\',\nthe instance object of the decorated method is passed to the decorator as the\nargument ``decorated_self``. If a dynamic decorator is not found at runtime\n(e.g., because it is a method of an optional class that has not been merged),\nthen the code execution proceeds normally, as shown below with the decorator\n``non_existent_decorator``:\n\n.. code:: python\n\n    from dyndesign import decoratewith\n\n    class Base:\n        def __init__(self):\n            self.comp = Component()\n\n        @decoratewith("comp.decorator1", "comp.decorator2", "non_existent_decorator")\n        def m(self):\n            print("Method `m` of class `Base`")\n\n    class Component:\n        def __init__(self):\n            self.value = "Initial"\n\n        def decorator1(self, func, decorated_self):\n            print(f"Beginning of method decoration #1 ({self.value=})")\n            self.value = "Processed"\n            func(decorated_self)\n            print("End of method decoration #1")\n\n        def decorator2(self, func, decorated_self):\n            print(f"Beginning of method decoration #2 ({self.value=})")\n            func(decorated_self)\n            print("End of method decoration #2")\n\n    base = Base()\n    base.m()\n\n    # Beginning of method decoration #1 (self.value=\'Initial\')\n    # Beginning of method decoration #2 (self.value=\'Processed\')\n    # Method `m` of class `Base`\n    # End of method decoration #2\n    # End of method decoration #1\n\n\nSafezone Context Manager\n------------------------\n\nAny function or method that may or may not exist at runtime (e.g., methods of\nmerged classes) can be invoked from Context Manager ``safezone`` in order to\nsuppress the possible exceptions raised if the function or method is not found\nat runtime. Optionally, a fallback function/method can be also passed. If no\nfunction name(s) is passed as argument of ``safezone``, then each function in\nthe safe zone\'s code is protected; if any function name(s) is passed, the\nprotection is restricted to the functions having that/those name(s). For\nexample, ``safezone`` can be used to safely call functions that may or may not\nexist at runtime:\n\n.. code:: python\n\n    from dyndesign import safezone\n    \n    def fallback():\n        print("Fallback function")\n\n    def function_a():\n        print("Function `a`")\n\n    with safezone(fallback=fallback):\n        function_a()\n        non_existent_function()\n\n    # Function `a`\n    # Fallback function\n\n\nA further example shows that ``safezone`` can be used to safely invoke methods\nof classes that may or may not be merged with other classes:\n\n.. code:: python\n\n    from dyndesign import safezone\n\n    class Base:\n        def fallback(self):\n            print("Fallback method")\n\n        def m(self, class_desc):\n            print(f"Method `m` of {class_desc}")\n            with safezone("optional_method", fallback=self.fallback):\n                self.optional_method()\n\n    class ExtOptional:\n        def optional_method(self):\n            print("Optional method from class `ExtOptional`")\n\n    merged = mergeclasses(Base, ExtOptional)()\n    merged.m("merged class")\n    base = Base()\n    base.m("class `Base` standalone")\n\n    # Method `m` of merged class\n    # Optional method from class `ExtOptional`\n    # Method `m` of class `Base` standalone\n    # Fallback method\n\n\nInvoking methods safely\n-----------------------\n\nAs alternative to ``safezone`` context manager, ``safeinvoke`` can be used to\nsafely invoke methods that may or may not exist at runtime. To this end, method\n``m`` of class ``Base`` of the example above can be replaced as follows:\n\n.. code:: python\n\n    from dyndesign import safeinvoke\n\n    ...\n\n        def m(self, class_desc):\n            print(f"Method `m` of {class_desc}")\n            safeinvoke("optional_method", self, fallback=self.fallback)\n\n\nSingleton classes\n-----------------\n\nSingleton classes can be swiftly created and destroyed:\n\n.. code:: python\n\n    from dyndesign import SingletonMeta\n\n    class Singleton(metaclass=SingletonMeta):\n        def __init__(self, instance_id = None):\n            if instance_id:\n                self.instance_id = instance_id\n            print(f"Created a {instance_id} instance of `Singleton`")\n\n        def where_points(self, object_name):\n            print(f"Object `{object_name}` points to the {self.instance_id} instance")\n\n    s_A = Singleton("first")\n    s_A.where_points("s_A")\n    s_B = Singleton()\n    s_B.where_points("s_B")\n    Singleton().destroy_singleton()\n    s_C = Singleton("second")\n    s_C.where_points("s_C")\n\n    # Created a first instance of `Singleton`\n    # Object `s_A` points to the first instance\n    # Object `s_B` points to the first instance\n    # Created a second instance of `Singleton`\n    # Object `s_C` points to the second instance\n\nThe class method ``destroy`` of SingletonMeta can be invoked to destroy the\nSingleton classes all at once. As a further alternative to the instance call\n``Singleton().destroy_singleton()``, the names of the Singleton classes to\ndestroy can be passed to the class method ``destroy``:\n\n.. code:: python\n\n    Singleton().destroy_singleton()\n    SingletonMeta.destroy()\n    SingletonMeta.destroy(\'Singleton1\', \'Singleton2\', \'Singleton3\')\n\n\nImporting classes dynamically\n-----------------------------\n\nClasses can be imported dynamically using the package/class names or the path in\ndot-notation as shown below:\n\n.. code:: python\n\n    from dyndesign import importclass\n\n    ClassA = importclass(\'package_A\', \'ClassA\')\n    ClassB = importclass(\'directory_B.package_B.ClassB\')\n\n\nRunning tests\n--------------\n\nTo run the tests using your default python:\n\n::\n\n    pip install -U pytest\n    python3 -m pytest test\n\n\n.. |Build Status| image:: https://github.com/amarula/dyndesign/actions/workflows/python-app.yml/badge.svg\n    :target: https://github.com/amarula/dyndesign/actions\n.. |Python Version Status| image:: https://img.shields.io/badge/python-3.8_3.9_3.10_3.11-blue.svg\n    :target: https://github.com/amarula/dyndesign/actions\n.. |PyPi Version Status| image:: https://badge.fury.io/py/dyndesign.svg\n    :target: https://badge.fury.io/py/dyndesign\n.. |License| image:: https://img.shields.io/badge/License-MIT-yellow.svg\n    :target: https://opensource.org/licenses/MIT\n',
    'author': 'Patrizio Gelosi',
    'author_email': 'patrizio.gelosi@amarulasolutions.com',
    'maintainer': 'Patrizio Gelosi',
    'maintainer_email': 'patrizio.gelosi@amarulasolutions.com',
    'url': 'https://github.com/amarula/dyndesign',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
